#!/usr/bin/env python3
"""
difpy_opt.py — super-fast, in-memory duplicate & similar image finder.
"""

import os, sys, glob, hashlib, logging, argparse, time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
import numpy as np
from PIL import Image
import numba
import imagehash

# ——— Logging —————————————————————————————————————————————————————————————————————————————
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)s: %(message)s",
    datefmt="%H:%M:%S",
)
logger = logging.getLogger(__name__)


# ——— Numba-accelerated MSE functions ————————————————————————————————————————————————————
@numba.njit
def _mse(a, b):
    """Compute mean squared error between a and b."""
    total = 0.0
    for i in range(a.shape[0]):
        diff = a[i] - b[i]
        total += diff * diff
    return total / a.shape[0]


@numba.njit
def _mse_thresh(a, b, thresh):
    """Compute MSE but bail out early if running mean exceeds thresh."""
    total = 0.0
    for i in range(a.shape[0]):
        diff = a[i] - b[i]
        total += diff * diff
        if (total / (i + 1)) > thresh:
            return total / (i + 1)
    return total / a.shape[0]


def compute_min_mse(a_flat: np.ndarray, b_rots: np.ndarray) -> float:
    """
    Return the minimum full MSE over all rotations in b_rots,
    using the pure _mse function.
    """
    best = 1e12
    for k in range(b_rots.shape[0]):
        mse = _mse(a_flat, b_rots[k])
        if mse < best:
            best = mse
    return best


# ——— Utilities ————————————————————————————————————————————————————————————————————————————
def md5sum(path: str) -> str:
    """Return MD5 checksum of a file."""
    h = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(8192), b""):
            h.update(chunk)
    return h.hexdigest()


def scan_files(dirs, recursive=True, exts=None):
    """
    Gather all files under dirs (list), optionally filtering by extensions tuple.
    """
    out = []
    for p in dirs:
        if os.path.isfile(p):
            out.append(p)
        elif os.path.isdir(p):
            pat = "**/*" if recursive else "*"
            out += glob.glob(os.path.join(p, pat), recursive=recursive)
    return [
        f for f in out if os.path.isfile(f) and (not exts or f.lower().endswith(exts))
    ]


def load_features(fp: str, px: int, bins: int, hash_size: int):
    """
    Open image, compute perceptual hash, resize to px×px, then
    produce:
      - orig size tuple,
      - flat array (float16),
      - 4×flat rotated arrays (float16),
      - perceptual hash string,
      - coarse histogram key (tuple).
    """
    try:
        img = Image.open(fp).convert("RGB")
        orig = img.size
        ph = str(imagehash.phash(img.resize((hash_size, hash_size))))
        img = img.resize((px, px), Image.BICUBIC)
        arr = np.asarray(img, np.uint8)
        flat = arr.ravel().astype(np.float16)
        rots = [
            np.asarray(np.rot90(arr, k)).ravel().astype(np.float16) for k in (1, 2, 3)
        ]
        rots_flat = np.vstack([flat] + rots)
        # build a 3×bins histogram, normalized to ints
        h = [np.histogram(arr[:, :, c], bins=bins, range=(0, 255))[0] for c in range(3)]
        hist = np.concatenate(h)
        key = tuple((hist * 100 // hist.sum()).tolist())
        return orig, flat, rots_flat, ph, key
    except Exception as e:
        logger.warning("Failed %s: %s", fp, e)
        return None


# ——— Core Finder —————————————————————————————————————————————————————————————————————————
class DuplicateFinder:
    def __init__(
        self,
        directories,
        recursive=True,
        px_size=50,
        hist_bins=8,
        similarity=0.0,
        threads=None,
        hash_size=16,
        vec_thresh=100,
    ):
        exts = (".jpg", ".jpeg", ".png")
        self.files = scan_files(directories, recursive, exts)
        self.px = px_size
        self.bins = hist_bins
        self.sim = float(similarity)
        self.threads = threads or os.cpu_count()
        self.hash_size = hash_size
        self.vec_thresh = vec_thresh

    def run(self):
        t0 = time.perf_counter()
        logger.info("Stage 1: exact-files (size→MD5)")
        size_map = defaultdict(list)
        for f in self.files:
            try:
                size_map[os.path.getsize(f)].append(f)
            except OSError:
                logger.warning("stat failed %s", f)

        exact, rem = [], []
        for grp in size_map.values():
            if len(grp) > 1:
                md5m = defaultdict(list)
                for f in grp:
                    md5m[md5sum(f)].append(f)
                for same in md5m.values():
                    if len(same) > 1:
                        exact.append(same)
                for f in grp:
                    if not any(f in g for g in exact):
                        rem.append(f)
            else:
                rem.extend(grp)

        logger.info("Stage 2: feature-extract via %d threads", self.threads)
        feats = {}
        with ThreadPoolExecutor(max_workers=self.threads) as ex:
            futures = {
                ex.submit(load_features, fp, self.px, self.bins, self.hash_size): fp
                for fp in rem
            }
            for fut in as_completed(futures):
                res = fut.result()
                if res:
                    feats[futures[fut]] = res

        logger.info("Stage 3: bucket by perceptual hash")
        buckets = defaultdict(list)
        for fp, (*_, ph, _) in feats.items():
            buckets[ph].append(fp)

        edges = []
        logger.info("Stage 4: within-pHash buckets")
        for fps in buckets.values():
            if len(fps) < 2:
                continue
            # further bucket by histogram key
            hist_b = defaultdict(list)
            for fp in fps:
                hist_b[feats[fp][4]].append(fp)
            for grp in hist_b.values():
                K = len(grp)
                if K < 2:
                    continue
                # small bucket + “similar” mode → early-abandon loops
                if self.sim > 0.0 and K <= self.vec_thresh:
                    for i in range(K):
                        A = grp[i]
                        flatA, rotsA = feats[A][1], feats[A][2]
                        for j in range(i + 1, K):
                            B = grp[j]
                            mse = min(
                                _mse_thresh(flatA, r, self.sim) for r in feats[B][2]
                            )
                            if mse <= self.sim:
                                edges.append((A, B, mse))
                else:
                    for i in range(K):
                        A = grp[i]
                        flatA = feats[A][1]
                        for j in range(i + 1, K):
                            B = grp[j]
                            if self.sim == 0.0:
                                if any(np.array_equal(flatA, r) for r in feats[B][2]):
                                    edges.append((A, B, 0.0))
                            else:
                                mse = compute_min_mse(flatA, feats[B][2])
                                if mse <= self.sim:
                                    edges.append((A, B, mse))

        logger.info("Stage 5: union-find grouping")
        parent = {}

        def find(x):
            parent.setdefault(x, x)
            if parent[x] != x:
                parent[x] = find(parent[x])
            return parent[x]

        def union(a, b):
            ra, rb = find(a), find(b)
            if ra != rb:
                parent[rb] = ra

        for A, B, _ in edges:
            union(A, B)

        groups = defaultdict(list)
        for fp in feats:
            groups[find(fp)].append(fp)
        for grp in exact:
            groups[grp[0]] = grp

        logger.info("Stage 6: format result")
        result, lower = {}, []
        for root, fps in groups.items():
            # pick highest-resolution as primary
            fps_sorted = sorted(
                fps, key=lambda f: feats[f][0][0] * feats[f][0][1], reverse=True
            )
            primary = fps_sorted[0]
            matches = []
            for other in fps_sorted[1:]:
                m = next((m for a, b, m in edges if {a, b} == {primary, other}), 0.0)
                matches.append([other, m])
                lower.append(other)
            result[primary] = matches

        stats = {
            "total_files": len(self.files),
            "featurized": len(feats),
            "groups": len(result),
            "duration_s": round(time.perf_counter() - t0, 3),
        }
        logger.info("Done %.3f s, groups=%d", stats["duration_s"], stats["groups"])
        return result, lower, stats


if __name__ == "__main__":
    p = argparse.ArgumentParser(description="difpy_opt: optimized finder")
    p.add_argument("-D", "--dirs", nargs="+", default=[os.getcwd()])
    p.add_argument("-r", "--recursive", action="store_true")
    p.add_argument("-px", "--px_size", type=int, default=50)
    p.add_argument("-b", "--bins", type=int, default=8)
    p.add_argument("-s", "--sim", type=float, default=0.0)
    p.add_argument("-t", "--threads", type=int, default=os.cpu_count())
    p.add_argument("--hash_size", type=int, default=16)
    p.add_argument("--vec_thresh", type=int, default=100)
    args = p.parse_args()

    finder = DuplicateFinder(
        args.dirs,
        args.recursive,
        args.px_size,
        args.bins,
        args.sim,
        args.threads,
        args.hash_size,
        args.vec_thresh,
    )
    res, low, stats = finder.run()
    print(
        __import__("json").dumps(
            {"result": res, "lower_quality": low, "stats": stats}, indent=2
        )
    )
